/***************************************************************************
  KNOCKING PUZZLE: the player hears a sequence of knocks. If they
  repeat the sequence of knocks correctly, the next sequence will play.
  The puzzle is solved when all sequences have been repeated correctly.
 ***************************************************************************
  Version: KnockingPuzzle V4_0
  Creation Date: January 2022
  Project: IWB Klopfrätsel  
  Original Author: Ruben Schär 
  Contact: ru.schaer@gmail.com
****************************************************************************
  FUNCTION: At boot, preset knock sequences are retreived from a txt file on
  the SD card. An external signal starts the game: a single "knock.mp3" is
  played repeatedly by the DFPlayerMini MP3 module, with short or long pauses
  as per SD card data. The sequence repeats every 10 seconds, until the player
  answers the challenge: a piezo transducer registers the players knocks.  
****************************************************************************/

#include <SPI.h>                  // required for SD card reader
#include <SD.h>                   // required for SD card reader
#include "SoftwareSerial.h"       // required for MP3 player module
#include "DFRobotDFPlayerMini.h"  // https://github.com/DFRobot/DFRobotDFPlayerMini/archive/1.0.3.zip

// GPIO mapping
#define gameStartPin 8  // input pin for "game start" signal
#define gameSolvedPin 7 // output pin for "game solved" signal
#define piezoPin 2      // input pin for piezo signal
#define spiCSpin 10     // spi chip select pin for SD card reader
#define txPin 4         // SofwareSerial TX pin
#define rxPin 3         // SofwareSerial RX pin
#define redLEDpin A0    // wrong sequence indicator
#define ylwLEDpin A1    // correct sequence indicator
#define grnLEDpin A2    // puzzle solved indicator

#define maxSequenceLenght 9 // configures presetSequence[array-columns] 
#define maxNrOfSequences 10  // configures presetSequence[array-rows] 
#define lenght 0            // default index for sequence lenght data
#define shortPause 420      // [ms] between knock playback for "short pause"
#define longPause 800       // [ms] between knock playback for "long pause"


// arrays, variables for knock sequence playback and detection
uint16_t presetSequence[maxNrOfSequences][maxSequenceLenght + 1]; // +1 for 0 index int value showing real sequence lenght
uint16_t playerSequence[maxSequenceLenght];                       // knock sequence entered by player

uint8_t currentSequenceIndex = 0;   // points to active preset sequence
uint8_t totalSequences = 0;         // holds number of preset sequences retreived from SD card
uint8_t currentSequenceLenght = 0;  // holds number of knocks (pauses) in current sequence

// isr variables
volatile bool isrFlag = 0;   // set by ISR event
volatile bool startGame = 1; // set by HIGH signal on gameStartPin

// timer variables
uint32_t oldMillis = 0; // used for long pause between sequences
uint32_t newMillis = 0; // used for long pause between sequences
uint32_t startTime = 0; // used to record time between knocks
uint32_t stopTime = 0;  // used to record time between knocks

// function declarations
void(* resetFunc) (void) = 0;//declare reset function at address 0
void setISR();          // reset all isr flags, attach interrupt
void unsetISR();        // reset all isr flags, detach interrupt
void ISRoutine();       // sets isrFlag = true and detaches interrupt
void playBackLoop(uint16_t timeLimit); // blocking loop, repeats active sequence every X ms until player input event

// class initializations
SoftwareSerial mySoftwareSerial(txPin, rxPin);
DFRobotDFPlayerMini myDFPlayer;
File sdFile;

//********************************************
//************SETUP START*********************
//********************************************

void setup()
{
  // USB serial out for debugging
  Serial.begin(9600);   
  Serial.println("Knocking Puzzle V4_0 19.01.2022");

  // set IO pinmodes
  pinMode(redLEDpin, OUTPUT);
  pinMode(grnLEDpin, OUTPUT);
  pinMode(ylwLEDpin, OUTPUT);
  pinMode(gameStartPin, INPUT_PULLUP);
  pinMode(gameSolvedPin, OUTPUT);
  digitalWrite(gameSolvedPin, HIGH); //default state is high (inverted by opto coupler)

  // init MP3 module serial connection
  mySoftwareSerial.begin(9600);
  delay(100);
  myDFPlayer.begin(mySoftwareSerial);
  myDFPlayer.volume(30);    // MP3 playback volume 0-30
  
  // init SD card, load data
  init_sdCard();
  loadPresetSequences();
  printPresetSequenceArray();
}

//********************************************
//************MAIN LOOP START*****************
//********************************************

void loop()
{
  unsetISR();
  
  while (startGame) // reset game, wait for start signal
  {
    currentSequenceIndex = 0;
    startGame = digitalRead(gameStartPin);
    delay(1);
  }
  digitalWrite(grnLEDpin, LOW);   // reset "solved" signal & indicator
  digitalWrite(gameSolvedPin, HIGH); 

  currentSequenceLenght = presetSequence[currentSequenceIndex][lenght]; // get lenght of active sequence

  playBackLoop(5000); // repeat active sequence while waiting for player knock

  if (recordPlayerSequence() == true) // upon completed player sequence...
  {


    if (checkPlayerSequenceVersus(currentSequenceIndex) == true) // compare player sequence against preset
    {
      currentSequenceIndex++;                        // increment currentSequenceIndex if player sequence matched
      if (currentSequenceIndex == (totalSequences))  // if the active sequence was the final sequence...
      {
        Serial.println("PUZZLE SOLVED");
        digitalWrite(gameSolvedPin, LOW);          // ...solve the puzzle and halt game until new start signal event (startGame = 0)
        currentSequenceIndex = 0;
        startGame = 1;
        digitalWrite(grnLEDpin, HIGH);
        digitalWrite(ylwLEDpin, HIGH);
        delay(1500); 
        digitalWrite(ylwLEDpin, LOW);
        delay(3500);
        digitalWrite(gameSolvedPin, HIGH);
        digitalWrite(grnLEDpin, LOW);        
      }

      else
      {
        Serial.println("CORRECT SEQUENCE");
        digitalWrite(ylwLEDpin, HIGH);
        delay(1500);  // wait x seconds before presenting NEXT sequence
        digitalWrite(ylwLEDpin, LOW);
      }
    }

    else
    {
      Serial.println("WRONG SEQUENCE");
      digitalWrite(redLEDpin, HIGH);
      delay(1500);  // wait x seconds before repeating CURRENT sequence
      digitalWrite(redLEDpin, LOW);
    }
  }
}
//********************************************
//**************MAIN LOOP END*****************
//********************************************


//********************************************
//**********Record Player Sequence************
//********************************************
//This function is called after an ISR event is triggered
//by the piezo input. The function records the time between each
//knock by the player. Once the expected number of knocks for the
//active sequence has been reached, the function returns "1".
//If the player input ends prematurely, the function returns "0"
//after a set amount of time has passed.

bool recordPlayerSequence()
{
  checkGameEnable();
  uint8_t playerKnockIndex = 0;
  for (uint8_t knock = 1; knock <= currentSequenceLenght; knock++)
  {
    startTime = millis();
    delay(100);
    setISR();
    while (!isrFlag)
    {
      if (wait(10000) == 1)
      {
        return 0; // return to main loop if player input times out
      }
    }
    unsetISR();
    stopTime = millis();

    playerSequence[playerKnockIndex] = (stopTime - startTime);
    playerKnockIndex++;
  }
  return 1;
}
//********************************************
//**********Check Player Sequence*************
//********************************************

// checkPlayerSequenceVersus(activeSequence) determines the shortest and longest pause
// in the players's sequence, and uses those values to create a relative
// evaluation of the pause durations, instead of using a predetermined value+tolerance.
// 
// The playeer sequence, initially a sequence of discrete millisecond values, is converted into
// a representative sequence of short and long pauses. That sequence is then compared
// to the active sequence in the presetSequence[array]. 
//
// The function returns 1 (correct) if the sequences match, and 0 (incorrect) if they don't.

bool checkPlayerSequenceVersus(int rIndex)
{
  checkGameEnable();
  uint16_t shortestTime = 15000;
  uint16_t longestTime = 0;
  uint16_t timeThreshold = 0;
  uint16_t checkSumm = 0;

  //the checksum determines if a sequence consist of one type of pause,
  //and sets a fixed time treshold of 500ms
  for (uint8_t i = 0; i < (currentSequenceLenght); i++) {
    checkSumm += presetSequence[currentSequenceIndex][i];
  }
  if (((uint16_t)checkSumm / currentSequenceLenght) == (shortPause || longPause)) {
    timeThreshold = 650;
  }

  else {
    for (uint8_t i = 0; i < (currentSequenceLenght); i++)
    {
      if (playerSequence[i] > longestTime)
      {
        longestTime = playerSequence[i];
      }
      if (playerSequence[i] < shortestTime)
      {
        shortestTime = playerSequence[i];
      }
    }

    timeThreshold = longestTime + shortestTime;
    timeThreshold = (uint16_t)timeThreshold / 2;
  }
  for (uint8_t i = 0; i <= (currentSequenceLenght - 1); i++)
  {
    if (playerSequence[i] > timeThreshold)
    {
      playerSequence[i] = longPause;
    }
    else
    {
      playerSequence[i] = shortPause;
    }
  }

  for (uint8_t i = 0; i <= (currentSequenceLenght - 1); i++)
  {
    if (playerSequence[i] != presetSequence[currentSequenceIndex][i + 1])
    {
      return false;
    }
    if (i == (currentSequenceLenght - 1))
    {
      return true;
    }
  }
}



//********************************************
// KNOCKING SOUND FUNCTIONS
// Generate repeating knocks from preset pattern data

void playBackLoop(uint16_t timeLimit)
{
  checkGameEnable();
  Serial.println("Playing active knock sequence");
  knockingSound();
  oldMillis = 0;

  while (!isrFlag)
  {
    if (oldMillis == 0)
    {
      oldMillis = millis();
    }

    newMillis = millis();
    if ((newMillis - oldMillis) > timeLimit)
    {
      Serial.println("Playing active knock sequence");
      knockingSound();
      oldMillis = 0;
    }
  }
  unsetISR();
}

//********************************************
void knockingSound() {
  unsetISR();
  for (uint8_t i = 1; i <= (currentSequenceLenght); i++)
  {
    checkGameEnable();
    myDFPlayer.play(2);
    delay(presetSequence[currentSequenceIndex][i]);
  }
  myDFPlayer.play(2);
  delay(100);
  setISR();
}

//********************************************
//wait() returns 1 after set amount of time, 0 if the timer
//is stil counting, and 2 if the isrFlag is set during wait

uint8_t wait(uint16_t timeLimit)
{
  checkGameEnable();
  if (oldMillis == 0)
  {
    oldMillis = millis();
  }

  newMillis = millis();
  if (newMillis - oldMillis > timeLimit)
  {
    oldMillis = 0;
    return 1;
  }

  if (isrFlag)
  {
    return 2;
  }

  else
  {
    return 0;
  }
}

//********************************************
// ISR HANDLING FUNCTIONS

void setISR()
{
  detachInterrupt(digitalPinToInterrupt(piezoPin));
  EIFR = 0x01; // manually reset the ISR register flag
  isrFlag = 0;
  delay(3);
  attachInterrupt(digitalPinToInterrupt(piezoPin), ISRoutine, FALLING);
  delay(3);
}
void unsetISR()
{
  detachInterrupt(digitalPinToInterrupt(piezoPin));
  EIFR = 0x01;
  isrFlag = 0;
}

void ISRoutine()
{
  detachInterrupt(digitalPinToInterrupt(piezoPin));
  EIFR = 0x01;
  isrFlag = 1;
}


//********************************************
// SD CARD HANDLING FUNCTIONS
// Init SD card

void init_sdCard()
{
  if (!SD.begin(spiCSpin))
  {
    Serial.println("sd card init failed");
    while (1)
      ;
  }
  Serial.println("sd card init done");
}

//********************************************
// loadPresetSequences() loads sequences from "pattern.txt" 
// file on the SD card and transfers the sequence data into the
// presetSequence[array]

void loadPresetSequences()
{
  uint8_t rIndex = 0;           // row index for 2D array
  uint8_t cIndex = 1;           // column dnex for 2D array (starts at index 1, 0 reserved for sequence lenght value)
  uint8_t knockCounter = 0;     // holds the number of knocks retreived in a row
  uint8_t patternCounter = 0;   // holds the numebr of sequences retreived from patterns.txt
  bool doubleX = 0;             // detects two 'X' in series, equals short pause

  sdFile = SD.open("patterns.txt");
  if (sdFile)
  {

    while (sdFile.available())
    {
      char c = sdFile.read();   // load next character  
      if (c == '\n')            // detect line feed character
      {
        if (knockCounter == 0)
        { // if knockCounter is 0 at this step, then the last row was terminated with alf. We treat that as EOF.
          sdFile.close();
          totalSequences = patternCounter;
          Serial.print(totalSequences);
          Serial.println(" sequences loaded.");
          doubleX = 0;
          cIndex = 0;
          rIndex = 0;          
          return;
        }
        
        else
        { // sequence complete, store current knockCounter as sequence lenght, advance to next row
          Serial.println(" ");
          presetSequence[rIndex][lenght] = knockCounter;
          rIndex++;
          patternCounter++;
          knockCounter = 0;
          doubleX = 0;
          cIndex = 1;
        }
      }

      else if ((c == '-')||(c == '-'))
      {
        Serial.print(c);
        presetSequence[rIndex][cIndex] = longPause;
        knockCounter++;
        cIndex++;
        doubleX = 0;
      }
      else if ((c == 'X')||(c == 'x'))
      {
        Serial.print(c);
        if (doubleX) {
          presetSequence[rIndex][cIndex] = shortPause;
          knockCounter++;
          cIndex++;          
        }
        else {
          doubleX = 1;
        }
      }
      else
      {
        doubleX = 0;
      }
    }
    sdFile.close();
    if (knockCounter > 0)
    { //if EOF is not preceded by a \n or \r, execute final incrementation and store values
      presetSequence[rIndex][0] = knockCounter;
      patternCounter++;
      totalSequences = patternCounter;
    }
    else
    {
      totalSequences = patternCounter;
    }
  }
  else
  {
    Serial.println("error opening file"); // if the file didn't open, print an error:
  }
}

// debug function: print all data contained in presetSequence[array]
void printPresetSequenceArray(){
  Serial.println(" ");
  Serial.println("[# of pauses]-[short pause=420ms]-[long pause=800ms]");
  uint8_t cIndex = 0;
  uint8_t rIndex = 0;
  for (rIndex = 0; rIndex < (totalSequences); rIndex++)
  {
    for (cIndex = 0; cIndex <= presetSequence[rIndex][0]; cIndex++)
    {
      Serial.print(presetSequence[rIndex][cIndex]);
      Serial.print(" ");
    }
    Serial.println(" ");    
  }
  Serial.println(" ");
  Serial.print("Number of sequences: ");
  Serial.println(totalSequences);
}

void checkGameEnable(){
  uint8_t GameEnable = digitalRead(gameStartPin);

  if(GameEnable){ resetFunc();}  else return;
  
}
